@@ -1,17 +1,28 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
+from sql import Column
 
+from trytond.config import config
+from trytond.i18n import gettext
 from trytond.model import Workflow, ModelView, fields
+from trytond.model.exceptions import AccessError
 from trytond.pool import PoolMeta, Pool
 from trytond.pyson import Eval, Bool, Id
 from trytond.transaction import Transaction
 from trytond.wizard import Wizard, StateTransition
 
-__all__ = ['PackageType', 'Package', 'ShipmentOut', 'CreateShipping']
+from .exceptions import PackWarning
+
+if config.getboolean('stock_package_shipping', 'filestore', default=False):
+    file_id = 'shipping_label_id'
+    store_prefix = config.get(
+        'stock_package_shipping', 'store_prefix', default=None)
+else:
+    file_id = store_prefix = None
 
 
-class PackageType(metaclass=PoolMeta):
-    __name__ = 'stock.package.type'
+class DimensionsMixin(object):
+    __slots__ = ()
 
     length = fields.Float('Length', digits=(16, Eval('length_digits', 2)),
         depends=['length_digits'])
@@ -78,14 +89,61 @@ class PackageType(metaclass=PoolMeta):
         return 2
 
 
-class Package(metaclass=PoolMeta):
+class PackageType(DimensionsMixin, metaclass=PoolMeta):
+    __name__ = 'stock.package.type'
+
+
+class Package(DimensionsMixin, metaclass=PoolMeta):
     __name__ = 'stock.package'
 
     shipping_reference = fields.Char('Shipping Reference',
         states={
             'readonly': Eval('_parent_shipment', {}).get('carrier', False),
             })
-    shipping_label = fields.Binary('Shipping Label', readonly=True)
+    shipping_label = fields.Binary(
+        "Shipping Label", readonly=True,
+        file_id=file_id, store_prefix=store_prefix)
+    shipping_label_id = fields.Char("Shipping Label ID", readonly=True)
+    shipping_tracking_url = fields.Function(
+        fields.Char(
+            "Shipping Tracking URL",
+            states={
+                'invisible': ~Eval('shipping_tracking_url'),
+                }),
+        'get_shipping_tracking_url')
+
+    def get_shipping_tracking_url(self, name):
+        return
+
+    @classmethod
+    def __register__(cls, module_name):
+        pool = Pool()
+        PackageType = pool.get('stock.package.type')
+        cursor = Transaction().connection.cursor()
+        table = cls.__table__()
+        package_type = PackageType.__table__()
+        table_h = cls.__table_handler__(module_name)
+        dimension_columns = [
+            'length', 'length_uom',
+            'height', 'height_uom',
+            'width', 'width_uom']
+        dimension_exists = any(
+            table_h.column_exist(c) for c in dimension_columns)
+
+        super().__register__(module_name)
+
+        # Migration from 5.8: Update dimensions on package from package_type
+        if not dimension_exists:
+            columns = []
+            values = []
+            for c in dimension_columns:
+                columns.append(Column(table, c))
+                values.append(package_type.select(
+                        Column(package_type, c),
+                        where=package_type.id == table.type))
+            cursor.execute(*table.update(
+                    columns=columns,
+                    values=values))
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -94,6 +152,17 @@ class Package(metaclass=PoolMeta):
             ('shipping_reference',) + tuple(clause[1:]),
             ]
 
+    @fields.depends('type')
+    def on_change_type(self):
+        super().on_change_type()
+        if self.type:
+            self.length = self.type.length
+            self.length_uom = self.type.length_uom
+            self.height = self.type.height
+            self.height_uom = self.type.height_uom
+            self.width = self.type.width
+            self.width_uom = self.type.width_uom
+
     @classmethod
     def copy(cls, packages, default=None):
         if default is None:
@@ -105,8 +174,7 @@ class Package(metaclass=PoolMeta):
         return super(Package, cls).copy(packages, default=default)
 
 
-class ShipmentOut(metaclass=PoolMeta):
-    __name__ = 'stock.shipment.out'
+class ShippingMixin:
 
     shipping_description = fields.Char('Shipping Description',
         states={
@@ -116,9 +184,7 @@ class ShipmentOut(metaclass=PoolMeta):
 
     @classmethod
     def __setup__(cls):
-        super(ShipmentOut, cls).__setup__()
-        # The shipment reference will be set by the shipping service
-        cls.reference.readonly = True
+        super().__setup__()
         cls._buttons.update({
                 'create_shipping': {
                     'invisible': (Eval('reference', False)
@@ -131,12 +197,66 @@ class ShipmentOut(metaclass=PoolMeta):
                         'root_packages'],
                     },
                 })
-        cls._error_messages.update({
-                'shipment_not_packed': ('The shipment "%(shipment)s" must be'
-                    ' packed before creating the shipping labels'),
-                'shipment_without_carrier': ('The shipment "%(shipment)s"'
-                    ' does not have a carrier set'),
-                })
+
+    @classmethod
+    def search_rec_name(cls, name, clause):
+        domain = super().search_rec_name(name, clause)
+        return ['OR', domain,
+            ('reference',) + tuple(clause[1:]),
+            ]
+
+    @classmethod
+    def validate(cls, shipments):
+        super().validate(shipments)
+        for shipment in shipments:
+            if shipment.carrier and shipment.carrier.shipping_service:
+                method_name = ('validate_packing_%s'
+                    % shipment.carrier.shipping_service)
+                validator = getattr(shipment, method_name)
+                validator()
+
+    @classmethod
+    def check_no_carrier(cls, shipments):
+        pool = Pool()
+        Warning = pool.get('res.user.warning')
+        for shipment in shipments:
+            if not shipment.carrier:
+                name = 'shipment_out_no_carrier_%s' % shipment
+                if Warning.check(name):
+                    raise PackWarning(name,
+                        gettext('stock_package_shipping'
+                            '.msg_shipment_without_carrier',
+                            shipment=shipment.rec_name))
+
+    @classmethod
+    @ModelView.button_action(
+        'stock_package_shipping.act_create_shipping_wizard')
+    def create_shipping(cls, shipments):
+        for shipment in shipments:
+            if shipment.state not in shipment.shipping_allowed:
+                raise AccessError(
+                    gettext('stock_package_shipping.msg_shipment_not_packed',
+                        shipment=shipment.rec_name))
+
+    @property
+    def shipping_allowed(self):
+        raise NotImplementedError
+
+    @property
+    def shipping_warehouse(self):
+        raise NotImplementedError
+
+    @property
+    def shipping_to(self):
+        raise NotImplementedError
+
+    @property
+    def shipping_to_address(self):
+        raise NotImplementedError
+
+
+class ShipmentOut(ShippingMixin, metaclass=PoolMeta):
+    __name__ = 'stock.shipment.out'
 
     @classmethod
     def __register__(cls, module):
@@ -155,47 +275,61 @@ class ShipmentOut(metaclass=PoolMeta):
                     & (model_data.model == 'ir.model.button')
                     & (model_data.fs_id == 'create_shipping_button')))
 
-    @classmethod
-    def search_rec_name(cls, name, clause):
-        domain = super(ShipmentOut, cls).search_rec_name(name, clause)
-        return ['OR', domain,
-            ('reference',) + tuple(clause[1:]),
-            ]
-
-    @classmethod
-    def validate(cls, shipments):
-        super(ShipmentOut, cls).validate(shipments)
-        for shipment in shipments:
-            if shipment.carrier and shipment.carrier.shipping_service:
-                method_name = ('validate_packing_%s'
-                    % shipment.carrier.shipping_service)
-                validator = getattr(shipment, method_name)
-                validator()
-
     @classmethod
     @Workflow.transition('packed')
     def pack(cls, shipments):
-        super(ShipmentOut, cls).pack(shipments)
-        for shipment in shipments:
-            if not shipment.carrier:
-                cls.raise_user_warning(
-                    'shipment_out_no_carrier_%s' % shipment.id,
-                    'shipment_without_carrier', {
-                        'shipment': shipment.rec_name,
-                        })
+        super().pack(shipments)
+        cls.check_no_carrier(shipments)
+
+    @property
+    def shipping_allowed(self):
+        return {'packed', 'done'}
+
+    @property
+    def shipping_warehouse(self):
+        return self.warehouse
+
+    @property
+    def shipping_to(self):
+        return self.customer
+
+    @property
+    def shipping_to_address(self):
+        return self.delivery_address
+
+
+class ShipmentInReturn(ShippingMixin, metaclass=PoolMeta):
+    __name__ = 'stock.shipment.in.return'
+
+    carrier = fields.Many2One(
+        'carrier', "Carrier",
+        states={
+            'readonly': ~Eval('state').in_(['draft', 'waiting', 'assigned']),
+            },
+        depends=['state'])
 
     @classmethod
-    @ModelView.button_action(
-        'stock_package_shipping.act_create_shipping_wizard')
-    def create_shipping(cls, shipments):
-        for shipment in shipments:
-            if shipment.state not in {'packed', 'done'}:
-                cls.raise_user_error('shipment_not_packed', {
-                        'shipment': shipment.rec_name,
-                        })
+    @ModelView.button
+    @Workflow.transition('done')
+    def done(cls, shipments):
+        super().done(shipments)
+        cls.check_no_carrier(shipments)
 
+    @property
+    def shipping_allowed(self):
+        return {'assigned', 'done'}
 
-# TODO Implement ShipmentInReturn
+    @property
+    def shipping_warehouse(self):
+        return self.from_location.warehouse
+
+    @property
+    def shipping_to(self):
+        return self.supplier
+
+    @property
+    def shipping_to_address(self):
+        return self.delivery_address
 
 
 class CreateShipping(Wizard):
@@ -205,4 +339,7 @@ class CreateShipping(Wizard):
     start = StateTransition()
 
     def transition_start(self):
+        shipping_service = self.record.carrier.shipping_service
+        method_name = 'validate_packing_%s' % shipping_service
+        getattr(self.record, method_name)()
         return 'end'
